/**
 ** David Bonnafous - Anthony Mallet - Wed Jul  4 2001
 ** Copyright (c) 2001 LAAS/CNRS 
 **
 ** $Log: p3d.c,v $
 ** Revision 1.4  2007/11/22 10:06:26  adegroot
 ** Add a workaround in case where the robot is not fully in the dtm.
 **
 ** If we can't compute the configuration of back axle, create a fake configuration
 ** based on the real position of the robot.
 **
 ** Revision 1.3  2007/09/10 10:55:23  sjoyeux
 ** return the found path instead than the first arc of the path
 **
 **   Instead of returning only the first arc, the search functions now assume
 **   that bestArc is an array if P3DArc structures of at least depth elements.
 **   This array is filled with the arcs composing the path, starting by the
 **   arc nearest to the goal.
 **
 **   The new output argument solutionDepth contains the number of arcs actually
 **   found in the solution path.
 **
 ** Revision 1.2  2007/01/08 19:55:17  sjoyeux
 ** * sync with the DEM tag (was something else before)
 ** * update to OpenRobot's T3Ds
 **
 ** Revision 1.1.1.1  2006/10/17 18:09:42  sjoyeux
 ** Packaged the P3D module from Lama's calife.
 ** See CHANGES for a list of modifications
 **
 **
 ** Revision 1.5.2.4  2005/04/13 20:27:38  tpeynot
 ** Modifications of protos and structures for body collision test (e.g. ground clearance info)
 **
 ** Revision 1.5.2.3  2005/04/07 20:42:45  tpeynot
 ** Add functions and elements in structures etc... for Dala body collision test
 ** Most of them are commented for the moment
 **
 ** Revision 1.5.2.2  2005/03/24 23:30:48  tpeynot
 ** adapt to map* and add print some useful info on pos in DEBUG mode
 **
 ** Revision 1.5.2.1  2005/03/18 23:42:38  tpeynot
 ** p3d uses for now an internal map: a struct P3D_MAP
 ** This struct will contain info a dem from the dem module
 **
 ** Revision 1.5  2005/03/17 13:46:53  tpeynot
 ** Add param "bestArc" ro p3dPlan fct to get info about the arc selectes by p3d planner
 ** Add functions adapted to Dala's structure instead of Lama. These functions are used if there's a flag -DATRV in Makefile. Otherwise, the lib will keep working with Lama adapted functions
 **
 ** Revision 1.4  2002/10/09 15:45:05  mallet
 ** p3dPlan retourne le cout estime, normalise entre 1 et +infini
 **
 ** Revision 1.3  2001/07/14 11:39:09  mallet
 ** premiere version ok
 **
 ** Revision 1.2  2001/07/12 17:38:45  mallet
 ** sync
 **
 ** Revision 1.1  2001/07/11 18:05:41  mallet
 ** re-ecriture de p3d
 **
 **/
#ident "$Id: p3d.c,v 1.4 2007/11/22 10:06:26 adegroot Exp $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/time.h>

#include "libp3d.h"
#include "p3dInt.h"

#define DEBUG 1

/*
 * p3dCreatePlanner -----------------------------------------------------
 *
 * Initialize a new planner
 */

P3DPlanner *
p3dCreatePlanner()
{
   P3DPlanner *p = malloc(sizeof(*p));

   if (p == NULL) {
      errno = ENOMEM;
      return NULL;
   }

   p->geom = NULL;
   p->robot = NULL;
   p->nArcs = 0;
   p->arcs = NULL;

   return p;
}

/*
 * p3dDestroyPlanner ----------------------------------------------------
 *
 * Free planner storage
 */

void
p3dDestroyPlanner(P3DPlanner *planner)
{
   if (planner == NULL) return;

   if (planner->geom != NULL) free(planner->geom);
   p3dRobotDestroy(planner->robot);
   p3dArcsDestroy(planner->arcs, planner->nArcs);
   free(planner);
}


/*
 * p3dConfigureGeom -----------------------------------------------------
 *
 * Create a geometric description.
 * This include the dimensions of the wheels and the arms and the maximum
 * allowed angles. unkMax is the maximum percentage of unknown terrain
 * under each wheel.
 */

int
p3dConfigureGeom(P3DPlanner *planner,
		 double len[3], double rad[2], double arm[2],
		 double bodyLength, double bodyWidth, 
		 double zBody, double groundClear,
		 double rollLimits[3], double pitchLimits[3], 
		 double alphaLimits[3], double betaLimits[3], double unkMax)
		 
{
   if (planner == NULL) return -1;
   if (planner->geom != NULL) free(planner->geom);

   planner->geom = malloc(sizeof(*planner->geom));
   if (planner->geom == NULL) { errno = ENOMEM; return -1; }

   planner->geom->len[0] = len[0];
   planner->geom->len[1] = len[1];
   planner->geom->len[2] = len[2];

   planner->geom->rad[0] = rad[0];
   planner->geom->rad[1] = rad[1];

   planner->geom->arm[0] = arm[0];
   planner->geom->arm[1] = arm[1];
   
   planner->geom->rollLimits[0] = rollLimits[0];
   planner->geom->rollLimits[1] = rollLimits[1];
   planner->geom->rollLimits[2] = rollLimits[2];

   planner->geom->pitchLimits[0] = pitchLimits[0];
   planner->geom->pitchLimits[1] = pitchLimits[1];
   planner->geom->pitchLimits[2] = pitchLimits[2];

   planner->geom->alphaLimits[0] = alphaLimits[0];
   planner->geom->alphaLimits[1] = alphaLimits[1];
   planner->geom->alphaLimits[2] = alphaLimits[2];

   planner->geom->betaLimits[0] = betaLimits[0];
   planner->geom->betaLimits[1] = betaLimits[1];
   planner->geom->betaLimits[2] = betaLimits[2];

   planner->geom->unkMax = unkMax;

#if (defined(ATRV) || defined(TROOPER)) || defined(RMP400)
   planner->geom->bodyLength = bodyLength;
   planner->geom->bodyWidth = bodyWidth;
   planner->geom->zBody = zBody;
   planner->geom->groundClear = groundClear;
#endif

   return 0;
}


/*
 * p3dConfigureRobot ----------------------------------------------------
 *
 * Create an internal robot description.
 * "scale" is the size of pixels in wheel bitmaps. This should be less
 * than the DEM resolution. 
 */

int
p3dConfigureRobot(P3DPlanner *planner, double scale, double rollStep, double vMin, double vMax,
									   double wMax, double accel, double decel)
{
   if (planner == NULL) { errno = EINVAL; return -1; }
   if (planner->geom == NULL) { 
      fputs("p3d: please configure geometry first!\n", stderr);
      errno = EINVAL; return -1;
   }

   if (planner->robot != NULL) p3dRobotDestroy(planner->robot);

   planner->robot = p3dCreateRobot(planner->geom, scale, rollStep, vMin, vMax, wMax, accel, decel);
   if (planner->robot == NULL) {
      fputs("p3d: cannot create robot\n", stderr);
      return -1;
   }


   return 0;
}

int		p3dConfigureRobotFromFile(P3DPlanner *planner, const char * name)
{
	FILE *fp = fopen(name, "r");
	if (fp == NULL) {
		fprintf(stderr, "%s : can't open %s (%d)\n", __func__, name, errno);
		return -1;
	}

	if (planner->robot != NULL) 
		p3dRobotDestroy(planner->robot);

	planner->robot = p3dReadP3DRobot(fp);
	fclose(fp);
	return 0;
}

int		p3dRobotSaveToFile(P3DPlanner *planner, const char *name)
{
	FILE *fp = fopen(name, "w");
	if (fp == NULL) {
		fprintf(stderr, "%s : can't open %s (%d)\n", __func__, name, errno);
		return -1;
	}

	if (planner->robot == NULL) {
		fprintf(stderr, "%s : can't save robot config, robot is NULL\n", __func__);
		fclose(fp);
		return -1;
	}

	p3dWriteP3DRobot(fp, planner->robot);
	fclose(fp);
	return 0;
}


/*
 * p3dConfigureArcs -----------------------------------------------------
 *
 * Configure the set of arcs to be evaluated by the planner.
 */

int
p3dConfigureArcs(P3DPlanner *planner, int nArcs, int nNodes, int depth,
		double rMin, double goalTol, double dMin)
{
	double distance;
	if (planner->geom == NULL) { 
		fputs("p3d: please configure geometry before configuring arcs !\n", stderr);
		errno = EINVAL; return -1;
	}


	if (planner->nArcs != 0) p3dArcsDestroy(planner->arcs, planner->nArcs);

	planner->goalTol = goalTol;
	planner->nArcs = 2*(nArcs/2)+1;
	planner->dMin = dMin;
	planner->arcs = p3dArcsCreate(nArcs, nNodes, dMin, 0.0, 0.0, 0.0,
			rMin, depth);
	if (planner->arcs == NULL) {
		fputs("p3d: cannot create arcs\n", stderr);
		return -1;
	}

	return 0;
}


/*
 * p3dPlan --------------------------------------------------------------
 *
 * Plan path on "map" with "planner", using robot position "mainToOrigin"
 * and trying to join goal "goalToOrigin" evaluating "nArcs" from "-rMin"
 * to "+rMin" on a distance between "dMin" and "dMax".
 *
 * Return 0 on success; "curv", "d" and "cost" represent the best arc.
 */

int
p3dPlan(P3DPlanner *planner, P3D_MAP *map,
	T3D *mainToOrigin, T3D *goalToOrigin,
	void (*callback)(P3DNode *node, void *data), void *data, 
	enum P3DWay way, P3DPlanSolution * sol)
{
   T3D mainToCenterOfMap, goalToCenterOfMap;
#if DEBUG
   int   k;
#endif
   int res = -1;
   struct timeval begin, end;
   double diff;
   double x,y,theta;

   gettimeofday(&begin, NULL);

   if (planner->robot == NULL) {
      fputs("p3d: robot not configured\n", stderr);
      return -1;
   }
   if (planner->arcs == NULL) {
      fputs("p3d: arcs not configured\n", stderr);
      return -1;
   }

   p3dComputeRobotPositionInMap(map, 
		   mainToOrigin, &mainToCenterOfMap, 
		   goalToOrigin, &goalToCenterOfMap);

#if DEBUG
   /*-- Print pos of Robot / origin of map --*/
   printf("p3dPlan: mainToCenterOfMap = ");
   for(k=0; k<6; k++)
     printf("%f ", mainToCenterOfMap.euler.euler[k]);
   printf("\n");

   /*-- Print pos of goal / origin of map --*/
   printf("p3dPlan: goalToCenterOfMap = ");
   for(k=0; k<6; k++)
     printf("%f ", goalToCenterOfMap.euler.euler[k]);
   printf("\n");
#endif 

   /* 
	* Compute the position that the robot will approximatvly have
	* when p3dPlan will be finised. For that, use the average time
	* and the previous speed orders
	*/

   x = mainToCenterOfMap.euler.euler[3];
   y = mainToCenterOfMap.euler.euler[4];
   theta = mainToCenterOfMap.euler.euler[0];
   
   p3dComputeRobotMovement(&x, &y, &theta, sol, planner);

   /*---- invoke search algorithm ----*/
	switch (way) {
		case P3D_ASTAR:
			res = p3dAStar(planner, map, 
					x, y, theta,						/* robot (x,y,theta) */
					goalToCenterOfMap.euler.euler[3],	/* goal (x,y) */
					goalToCenterOfMap.euler.euler[4],
					planner->goalTol,
					callback, data, sol);
			break;
		case P3D_FULL:
			res = p3dFull(planner, map,
					x, y, theta,						/* robot (x,y,theta) */
					goalToCenterOfMap.euler.euler[3],	/* goal (x,y) */
					goalToCenterOfMap.euler.euler[4],
					planner->goalTol,
					callback, data, sol);
			break;
	}

   gettimeofday(&end, NULL);

   diff = sub_timeval(&begin, &end);
   sol->average_time = ( sol->average_time * sol->ncalls + diff ) / (sol->ncalls+1);
   sol->ncalls++;

   return res;
}





/*
 * p3dCreateRobotConfig -----------------------------------------------------
 *
 * Initialize a new planner
 */

P3DRobotConfig *
p3dCreateRobotConfig()
{
   P3DRobotConfig *config = malloc(sizeof(*config));

   if (config == NULL) {
      errno = ENOMEM;
      return NULL;
   }

   config->q = 0.0;
   config->pUnk = 1.0;

   return config;
}

/*
 * p3dDestroyRobotConfig ----------------------------------------------------
 *
 * Free Robot Config
 */

void
p3dDestroyRobotConfig(P3DRobotConfig *config)
{
   if (config == NULL) return;

   free(config);
}



/*
 * p3dInfoP3DRobotConfig ----------------------------------------------------
 *
 */

void
p3dInfoP3DRobotConfig(P3DRobotConfig*  config)
{
  fprintf(stderr,"P3DRobotConfig:---------------------------------\n");
  fprintf(stderr,"config->q = %lf\n", config->q);
  fprintf(stderr,"config->x.y.z = %lf %lf %lf\n", config->x, config->y, config->z);
  fprintf(stderr,"dzBodyGround: %lf\n", config->dzBodyGround);
  fprintf(stderr,"roll(deg): %lf pitch(deg) = %lf\n", 
	  RAD_TO_DEG(config->roll), RAD_TO_DEG(config->pitch) );
  fprintf(stderr,"alpha1(deg): %lf alpha2 = %lf\n", 
	  RAD_TO_DEG(config->alpha1), RAD_TO_DEG(config->alpha2) );
  fprintf(stderr,"pUnk: %lf\n", config->pUnk);
  fprintf(stderr,"---------------------------------------------\n");
  
}
