/*
 * Copyright (C) 2001 LAAS/CNRS 
 *
 * Anthony Mallet - Thu Jul  5 2001
 *
 * $Log: p3dInt.h,v $
 * Revision 1.3  2007/09/10 10:55:23  sjoyeux
 * return the found path instead than the first arc of the path
 *
 *   Instead of returning only the first arc, the search functions now assume
 *   that bestArc is an array if P3DArc structures of at least depth elements.
 *   This array is filled with the arcs composing the path, starting by the
 *   arc nearest to the goal.
 *
 *   The new output argument solutionDepth contains the number of arcs actually
 *   found in the solution path.
 *
 * Revision 1.2  2007/01/08 19:55:17  sjoyeux
 * * sync with the DEM tag (was something else before)
 * * update to OpenRobot's T3Ds
 *
 * Revision 1.1.1.1  2006/10/17 18:09:42  sjoyeux
 * Packaged the P3D module from Lama's calife.
 * See CHANGES for a list of modifications
 *
 *
 * Revision 1.6.2.7  2005/07/13 15:54:23  tpeynot
 * three functions protos added
 *
 * Revision 1.6.2.6  2005/04/26 12:29:07  tpeynot
 * Add proto of function p3dInfoP3DPlanner (Info for Calife variable P3DPlanner)
 *
 * Revision 1.6.2.5  2005/04/21 12:41:41  tpeynot
 * put useful macros RAD_TO_DEG and DEG_TO_RAD as they can be used in several codels
 *
 * Revision 1.6.2.4  2005/04/21 12:31:30  tpeynot
 * Add gdheDriver.c, containing functions for gdhe connection
 *
 * Revision 1.6.2.3  2005/04/20 16:51:21  tpeynot
 * remove p3dPlaceRobotDala proto, now in libp3d.h
 *
 * Revision 1.6.2.2  2005/04/07 20:42:45  tpeynot
 * Add functions and elements in structures etc... for Dala body collision test
 * Most of them are commented for the moment
 *
 * Revision 1.6.2.1  2005/03/18 23:42:38  tpeynot
 * p3d uses for now an internal map: a struct P3D_MAP
 * This struct will contain info a dem from the dem module
 *
 * Revision 1.6  2005/03/17 13:46:53  tpeynot
 * Add param "bestArc" ro p3dPlan fct to get info about the arc selectes by p3d planner
 * Add functions adapted to Dala's structure instead of Lama. These functions are used if there's a flag -DATRV in Makefile. Otherwise, the lib will keep working with Lama adapted functions
 *
 * Revision 1.5  2002/10/09 15:45:05  mallet
 * p3dPlan retourne le cout estime, normalise entre 1 et +infini
 *
 * Revision 1.4  2001/07/18 19:46:03  mallet
 * printfs et finitions
 *
 * Revision 1.3  2001/07/14 11:39:09  mallet
 * premiere version ok
 *
 * Revision 1.2  2001/07/12 17:38:45  mallet
 * sync
 *
 * Revision 1.1  2001/07/11 18:05:41  mallet
 * re-ecriture de p3d
 *
 */

#ifndef P3DINT_H
#define P3DINT_H

#include <sys/time.h>

/*---- Useful Macros ----------*/
#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif
#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif
#ifndef SIGN
#define SIGN(x) ((x)<0.0?-1:1)
#endif
#ifndef SQ
#define SQ(x) ((x)*(x))
#endif

#ifndef RAD_TO_DEG
#define RAD_TO_DEG(x) ( (x)*180.0/M_PI )
#endif

#ifndef DEG_TO_RAD
#define DEG_TO_RAD(x) ( (x)*M_PI/180.0 )
#endif

/*---- Useful Constants ----------*/
#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif
#ifndef M_PI_2
#define M_PI_2          1.57079632679489661923
#endif
#ifndef M_PI_4
#define M_PI_4          .78539816339744830961
#endif

/*--------- Internal Functions protos ------------*/

P3DRobot *	p3dCreateRobot(P3DRobotGeometry *geom, double scale,
			double rollStep, double vMin, double vMax, double wMax, double accel, double decel);
P3DRobot*	p3dReadP3DRobot(FILE * fp);
void		p3dWriteP3DRobot(FILE *fp, const P3DRobot *robot);
void		p3dRobotDestroy(P3DRobot *robot);
double		p3dEvalConfig(P3DRobotGeometry *geom,
			P3DRobotConfig *current, P3DRobotConfig *prev);
P3DElement *	p3dLamaWheelCreate(double len[3], double rad[2],
			double roll, double scale);
void		p3dP3dElementDestroy(P3DElement *wheel);

P3DArc *	p3dArcsCreate(int nArcs, int nNodes, double distance, double x, double y,
			double theta, double rMin, int depth);
void		p3dArcsDestroy(P3DArc *arcs, int n);
void		p3dArcsResetFlags(P3DArc *arcs, int n);
void		p3dArcsReset(P3DArc *arcs, int n, double x, double y,
			double theta, double gx, double gy, double goalTol);
void		p3dArcComputeNodePosition(P3DArc *arcs, int n, double x, double y, 
									  double theta, double stop_distance);

double		p3dPlaceRobot(P3DRobot *robot, P3DRobotGeometry *geom,
			double x, double y, double theta,
			P3D_MAP *map, P3DRobotConfig *config);
double		p3dDubbins(double gx, double gy, double x, double y,
			double theta, double r);
void		p3dGetArcPos(double r, double ds, double x0, double y0,
			double theta0, double *x, double *y, double *theta);

int		p3dAStar(P3DPlanner *planner, P3D_MAP *map, double x,
			double y, double theta, double gx, double gy,
			double goalTol,
			void (*callback)(P3DNode *node, void *data),
			void *data, P3DPlanSolution *sol);

int		p3dFull(P3DPlanner *planner, P3D_MAP *map, double x,
			double y, double theta, double gx, double gy,
			double goalTol,
			void (*callback)(P3DNode *node, void *data),
			void *data, P3DPlanSolution *sol);

/*************************************/

void p3dInfoP3DPlanner(P3DPlanner*  planner);

double  p3dPlaceRobotDala(P3DRobot *robot, P3DRobotGeometry *geom,
			  double x, double y, double theta,
			  P3D_MAP *map, P3DRobotConfig *config);

double	p3dEvalConfigDala(P3DRobotGeometry *geom,
			  P3DRobotConfig *current, P3DRobotConfig *prev);


P3DElement *  p3dDalaWheelCreate(double len[3], double rad[2],
				    double roll, double scale);

P3DElement *  p3dDalaBodyCreate(double length, double width, double zBody, 
				double roll, double pitch, double scale);

/**********************************************************************/

void p3dInfoP3dMap(P3D_MAP*  map);

/**********************************************************************/

double sub_timeval(struct timeval *begin, struct timeval *end);

/*********************************************************************/
double  p3dPlaceRobotTrooper(P3DRobot *robot, P3DRobotGeometry *geom,
			  double x, double y, double theta,
			  P3D_MAP *map, P3DRobotConfig *config);


double	p3dEvalConfigTrooper(P3DRobotGeometry *geom,
			  P3DRobotConfig *current, P3DRobotConfig *prev);

/*********************************************************************/

#if defined(ATRV) || defined(RMP400)
#define PLACE_ROBOT(robot,geom,x,y,theta,map,config) \
		p3dPlaceRobotDala((robot), (geom), (x), (y), (theta), (map), (config))
#define EVAL_CONFIG(geom, current, prev) \
		p3dEvalConfigDala((geom),(current),(prev))
#elif defined(TROOPER)
#define PLACE_ROBOT(robot,geom,x,y,theta,map,config) \
		p3dPlaceRobotTrooper((robot), (geom), (x), (y), (theta), (map), (config))
#define EVAL_CONFIG(geom, current, prev) \
		p3dEvalConfigTrooper((geom),(current),(prev))
#else /* Lama case */
#define PLACE_ROBOT(robot,geom,x,y,theta,map,config) \
		p3dPlaceRobot((robot), (geom), (x), (y), (theta), (map), (config))
#define EVAL_CONFIG(geom, current, prev) \
		p3dEvalConfig((geom),(current),(prev))
#endif

/**********************************************************************/

double p3dGetStopDistance(double v, P3DPlanner *planner);
double max_speed(const P3DArc *a, const P3DPlanner *planner);
void p3dComputeRobotMovement(double *px, double *py, double *ptheta,
		P3DPlanSolution *sol, P3DPlanner * planner);
void check_unknow(P3DRobotGeometry *geom, P3DArc *a, int current_node);

/**********************************************************************/
void p3dComputeRobotPositionInMap(P3D_MAP * map, T3D *mainToOrigin, 
		T3D *mainToCenterOfMap, T3D *goalToOrigin, T3D *goalToCenterOfMap);

/**********************************************************************/


void p3dResetPlanSolution(P3DPlanSolution *sol);
void p3dSetPlanSolution(P3DPlanSolution *sol, P3DPlanner *planner, P3DArc *best);
		
/**********************************************************************/
double p3dGetVehiculeSize(P3DPlanner *planner);
double p3dGetDistanceFixAndMobileAxle(P3DPlanner *planner);

#endif /* P3DINT_H */
