/**
 ** Thierry Peynot - Mon Mar 21 2005
 ** Copyright (c) 2005 LAAS/CNRS 
 **
 ** Useful functions to manage P3D_MAP struct
 **
 **/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
/* for memset */
#include <string.h>

#include "libp3d.h"
#include "p3dInt.h"

#include <math.h>
#include <time.h>
#include <sys/time.h>

#include <gdhe/gdhe_terrain.h>

#define DEBUG 0


/**********************************************************************
 *
 * Creation and allocation of a P3D_MAP
 *
 **********************************************************************/

P3D_MAP*	
create_p3dmap ()

{
  P3D_MAP*   p3dmap;
  int        i,j;
  
  p3dmap = (P3D_MAP*) malloc (sizeof(P3D_MAP));
  
  if (p3dmap == NULL) {
	fprintf(stderr, "%s : Can't allocate memory for p3dMap\n", __func__);   
    return (NULL);
  }

  /* Initializes parameters */
  p3dmap->nbLines = 0;
  p3dmap->nbCols = 0;
  
  p3dmap->scale = 0.0;
  p3dmap->zScale = 0.0;

  p3dmap->xOrigin = 0.0;
  p3dmap->yOrigin = 0.0;
  p3dmap->zOrigin = 0.0;
  
  /** Init the map mainToOrigin as a T3DEuler **/
  p3dmap->mainToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);

  
  /* Default: all cells EMPTY */
  for(i = 0; i < P3D_MAP_MAX_LINES; i++)
   {
     for(j = 0; j < P3D_MAP_MAX_COLUMNS; j++)
       {
	 p3dmap->cells[i][j].state = P3D_MAP_CELL_EMPTY;
	 p3dmap->cells[i][j].z = 0.0;
       }
   }
  

  return(p3dmap);
}




/**********************************************************************
 *
 * Destruction of a P3D_MAP
 *
 **********************************************************************/

void
destroy_p3dmap (P3D_MAP*  p3dMap)
{
  
  if(p3dMap->mainToOrigin != NULL)
    t3dDelete(p3dMap->mainToOrigin);

  free(p3dMap);

}




/**********************************************************************
 *
 * Writing a P3D_MAP on a file
 *
 **********************************************************************/

int 
write_p3dmap(const P3D_MAP* map, const char* fileName)
{
  FILE*  file;
  int    i,j;

  file = fopen(fileName, "w");
  if (file == NULL) {
	  fprintf(stderr, "p3d: failed to open file %s : %s\n", fileName, strerror(errno));
	  return -1;
  }

  /* Write the file heading */
  fprintf(file, "%d\t%d\n", map->nbLines, map->nbCols);
  fprintf(file, "%f\t%f\n", map->xOrigin, map->yOrigin);
  for(i=0; i<6; i++)
	  fprintf(file, "%lf ", map->mainToOrigin->euler.euler[i]);
  fprintf(file, "\n");
  fprintf(file, "%f\n", map->scale);
  fprintf(file, "%f\t%f\n", map->zOrigin, map->zScale);

  /* print the cells in file */
  for(i = 0; i < map->nbLines; i++)
	  for(j = 0; j < map->nbCols; j++)
		  fprintf(file, "%d\t%f\t\t%f\n", map->cells[i][j].state, 
											map->cells[i][j].z, 
											map->cells[i][j].sigma_z);

  fclose(file);

  return 0;
}

int
read_p3dmap(P3D_MAP* map, const char* filename)
{
	FILE * f;
	int i,j;
	int nbLines, nbCols;

	f = fopen(filename, "r");
	if (f == NULL) {
		fprintf(stderr, "p3d: failed to open file %s : %s\n", filename, strerror(errno));
		return -1;
	}

	fscanf(f, "%d\t%d\n", &nbLines, &nbCols);
	if (nbLines > P3D_MAP_MAX_LINES || nbCols > P3D_MAP_MAX_COLUMNS) {
		fprintf(stderr, "p3d : p3dMap is too big\n");
		fclose(f);
		return -1;
	}

	map->nbLines = nbLines;
	map->nbCols = nbCols;

	fscanf(f, "%f\t%f\n", &map->xOrigin, &map->yOrigin);
	for(i=0; i<6; i++)
		fscanf(f, " %lf ", &map->mainToOrigin->euler.euler[i]);
	fscanf(f, "\n");
	fscanf(f, "%f\n", &map->scale);
	fscanf(f, "%f\t%f\n", &map->zOrigin, &map->zScale);

	for(i = 0; i < map->nbLines; i++)
		for(j = 0; j < map->nbCols; j++)
			fscanf(f, "%d\t%f\t%f\n", 
					&map->cells[i][j].state, 
					&map->cells[i][j].z, 
					&map->cells[i][j].sigma_z);
	fclose(f);

	return 0;
}

int
p3dmap_to_gdhe(const P3D_MAP* map, const char* filename)
{
	GDHE_MNT_HEADER hdr;
	FILE* f;
	int i, j, nbCells;
	GDHE_MNT_POINT* pts, *ppts = NULL;

	hdr.x0 = map->xOrigin;
	hdr.y0 = map->yOrigin;
	hdr.z0 = map->zOrigin;
	hdr.dx = map->scale;
	hdr.dy = map->scale;
	hdr.dz = map->zScale;
	hdr.storageType = GDHE_MNT_X_ROW;
	hdr.nrows = map->nbLines;
	hdr.ncols = map->nbCols;

	f = fopen(filename, "w");
	if (f == NULL) {
		fprintf(stderr, "p3d: %s : failed to open file %s : %s\n", __func__, 
				filename, strerror(errno));
		return -1;
	}

	pts = malloc(map->nbLines * map->nbCols * sizeof(GDHE_MNT_POINT));
	if (pts == NULL) {
		fprintf(stderr, "p3d: %s : failed to allocate memory\n", __func__);
		fclose(f);
		return -1;
	}

	ppts = pts;
	for (i = 0; i < map->nbLines; i++)
		for (j = 0; j < map->nbCols; j++) {
			if (map->cells[i][j].state == P3D_MAP_CELL_EMPTY) {
				ppts->z = 0;
				ppts->type = 0;
			} else {
				ppts->type = 1;
				ppts->z = round(((map->cells[i][j].z - map->zOrigin) / map->zScale));
			}
			ppts++;
		}


	if (fwrite (&hdr, sizeof (GDHE_MNT_HEADER), 1, f) != 1) {
		free(pts);
		fclose(f);
		return -1;
	}

	nbCells = hdr.nrows * hdr.ncols;

	if (fwrite (pts, sizeof (GDHE_MNT_POINT), nbCells, f) != nbCells) {
		free(pts);
		fclose(f);
		return -1;
	}

	free(pts);
	fclose(f);
	return 0;
}

/**********************************************************************
 *
 * Info about a P3D_MAP
 *
 **********************************************************************/

void
p3dInfoP3dMap(P3D_MAP*  map)
{
  printf("-------- p3dMap Infos --------------------------------\n");
  printf("map->nbLines    = %d\n", map->nbLines);
  printf("map->nbCols     = %d\n", map->nbCols);
  printf("map->xOrigin    = %f\n", map->xOrigin);
  printf("map->yOrigin    = %f\n", map->yOrigin);
  printf("map->scale      = %f\n",  map->scale);
  printf("map->zOrigin    = %f\n",  map->zOrigin);
  printf("map->zScale     = %f\n",  map->zScale);

}






/**********************************************************************
 *
 * Info about a P3DPlanner
 *  for the moment, prints the info about the firts arc (bestArc)
 *
 **********************************************************************/

void
p3dInfoP3DPlanner(P3DPlanner*  planner)
{
  P3DArc*  arc;
  int k;

  printf("-------- planner->arcs Infos --------------------------------\n");
  printf("planner->arcs.r       = %lf\n", planner->arcs->r);
  printf("planner->arcs.ds      = %lf\n", planner->arcs->ds);
  printf("planner->arcs.cost    = %lf\n", planner->arcs->cost);
  printf("planner->arcs.n       = %d\n",  planner->arcs->n);
  printf("planner->arcs.current = %d\n",  planner->arcs->current);

  arc = planner->arcs;
  for(k=0; k<arc->n; k++)
    {
      printf("-------- Node[%d] ------------------\n", k);
      printf("planner->arcs->nodes[%d].conf.q = %lf\n", k, arc->nodes[k].conf.q);
      printf("planner->arcs->nodes[%d].conf.x = %lf\n", k, arc->nodes[k].conf.x);
      printf("planner->arcs->nodes[%d].conf.y = %lf\n", k, arc->nodes[k].conf.y);
      printf("planner->arcs->nodes[%d].conf.z = %lf\n", k, arc->nodes[k].conf.z);
      printf("planner->arcs->nodes[%d].conf.dzBodyGround = %lf\n", k, arc->nodes[k].conf.dzBodyGround);
      printf("planner->arcs->nodes[%d].conf.roll   = %lf\n", k, arc->nodes[k].conf.roll);
      printf("planner->arcs->nodes[%d].conf.pitch  = %lf\n", k, arc->nodes[k].conf.pitch);
      printf("planner->arcs->nodes[%d].conf.alpha1 = %lf\n", k, arc->nodes[k].conf.alpha1);
      printf("planner->arcs->nodes[%d].conf.alpha2 = %lf\n", k, arc->nodes[k].conf.alpha2);
      printf("planner->arcs->nodes[%d].conf.beta3  = %lf\n", k, arc->nodes[k].conf.beta3);
      printf("planner->arcs->nodes[%d].conf.pUnk   = %lf\n", k, arc->nodes[k].conf.pUnk);
      printf("planner->arcs->nodes[%d].cost  = %lf\n", k, arc->nodes[k].cost);
    }
  
}






/**********************************************************************
 *
 * Try to Place the Robot on a given p3dMap, at a given location
 *
 **********************************************************************/

int
p3dPlaceRobotOnMap(P3DRobotConfig*  config, /* output: predicted config. */
		   P3DPlanner* p3dplanner,
		   double robot_x, double robot_y, /* Robot position in global frame */
		   double robot_thetaDeg,   /* Robot Theta in degrees */
		   P3D_MAP*   p3dMap,
		   int    display)  /* Display result in GDHE if display ==1 */
/* display ==0 => Print predicted config */
/* display >= 2 => Do not print or display anything */
{
  double    quality;
  T3D mainToCenterOfMap;
  T3D* mainToOrigin;
  double  robot_theta;
  double  x,y,theta; /* Robot position in map */
#if DEBUG
  int  i,j;
  P3DElement* body;
#endif

  if( (p3dplanner->robot == NULL) || (p3dplanner->geom == NULL) )
    {

      fputs("p3dPlaceRobotOnMap: Cannot place robot: No robot configured\n", stderr);
      return -1;
    }

#if DEBUG  /* Test: check all body bitmaps */
  for(i = 0; i < p3dplanner->robot->nBodyRoll; i++)
    {
      for(j = 0; j < p3dplanner->robot->nBodyPitch; j++)
	 {
	   body = p3dplanner->robot->body[i][j];
	   fprintf(stderr, "robot->body[%d][%d]: x = %f y = %f h = %f\n",
		   i, j, body->x, body->y, *(body->height) );
	 }
    }
#endif

  robot_theta = DEG_TO_RAD(robot_thetaDeg);

  mainToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  mainToOrigin->euler.euler[0] = robot_theta;
  mainToOrigin->euler.euler[3] = robot_x;
  mainToOrigin->euler.euler[4] = robot_y;

  p3dComputeRobotPositionInMap(p3dMap, mainToOrigin, &mainToCenterOfMap, NULL, NULL);
  
   x = mainToCenterOfMap.euler.euler[3];
   y = mainToCenterOfMap.euler.euler[4];
   theta = mainToCenterOfMap.euler.euler[0];
   
   /* Check that given position is in the map */
   if( (x >= p3dMap->nbLines*p3dMap->scale) || (y >= p3dMap->nbCols*p3dMap->scale))
     {

       fputs("p3dPlaceRobotOnMap: Cannot place robot: Given position is out of map\n", stderr);
       fprintf(stderr, "p3dPlaceRobotOnMap: Position: %lf %lf %lf\n", x,y,theta);
       return -1;
     }

#if DEBUG
   /* To Check */
   fprintf(stderr, "p3dPlaceRobotOnMap: executing p3dPlaceRobotDala(x=%lf, y=%lf, theta=%lf)... \n", x, y, theta);
#endif /* DEBUG */

   quality = PLACE_ROBOT(p3dplanner->robot, p3dplanner->geom, x, y, theta, p3dMap, config);

   /* If config unacceptable for the placeRobot function, show on which (global) position */
/*    if(quality < 0) */
/*      fprintf(stderr, "Global Position: x= %f, y= %f\n", robot_x, robot_y); */

   if(display < 2)
     {
       fprintf(stderr,"p3dPlaceRobot:---------------------------------\n");
       fprintf(stderr,"p3dPlaceRobot: Config at position %lf %lf %lf\n", robot_x, robot_y, RAD_TO_DEG(robot_theta) );
       fprintf(stderr,"p3dPlaceRobot: quality = %lf, config->q = %lf\n", quality, config->q);
       fprintf(stderr,"p3dPlaceRobot: config->x.y.z = %lf %lf %lf\n", config->x, config->y, config->z);
       fprintf(stderr,"p3dPlaceRobot: dzBodyGround: %lf\n", config->dzBodyGround);
       fprintf(stderr,"p3dPlaceRobot: roll(deg): %lf pitch(deg) = %lf\n", 
	       RAD_TO_DEG(config->roll), RAD_TO_DEG(config->pitch) );
       fprintf(stderr,"p3dPlaceRobot: sigmaPitch: %lf sigmaRoll = %lf sigma_z = %lf\n", 
	       config->sigmaPitch, config->sigmaRoll, config->sigmaz );
       fprintf(stderr,"p3dPlaceRobot: alpha1(deg): %lf alpha2(deg) = %lf\n", 
	       RAD_TO_DEG(config->alpha1), RAD_TO_DEG(config->alpha2) );
       fprintf(stderr,"p3dPlaceRobot: pUnk: %lf\n", config->pUnk);
   fprintf(stderr,"p3dPlaceRobot:---------------------------------\n");
     }
   

   t3dDelete(mainToOrigin);
   return 0;
}






/**********************************************************************
 *
 * Test Place Robot on a complete map
 *
 **********************************************************************/

int
p3dPlaceRobotOnWholeMap(P3DPlanner* p3dplanner,
			double thetaDeg,   /* Robot Theta in degrees */
			P3D_MAP*   map,
			double    x_min,
			double    y_min,
			double    x_max,
			double    y_max )
{
  P3DRobotConfig  config;
  double x,y;
  int    minLineNb;
  int    minColNb;
  int    maxLineNb;
  int    maxColNb;
  int  i, j;
  /* Struct for time */
  struct timespec tp;
  unsigned long msec;
  double timesec0;  /* Initial Time in seconds */
  double timesec1;  /* Final Time in seconds */
  /*****************/

  minLineNb = (int) ( (x_min-map->xOrigin) / map->scale );
  minColNb  = (int) ( (y_min-map->yOrigin) / map->scale );
  maxLineNb = (int) ( (x_max-map->xOrigin) / map->scale );
  maxColNb  = (int) ( (y_max-map->yOrigin) / map->scale );

  /* Check that max Line and Col numbers are inside the map */
  if ( (maxLineNb > map->nbLines) || (maxColNb > map->nbCols) )
    {
      fprintf(stderr, "%s : maxLineNb or maxColNb out of limits\n", __func__);
      return -1;
    }

  /* TODO: the min should be checked also... */

  /*** Initial Time ***/
  clock_gettime(CLOCK_REALTIME, &tp);
  msec = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
  timesec0  = ((double)(msec))/1000;
  /******************/

  x = map->xOrigin + ((float) minLineNb) * map->scale;

  /* Parcours de toute la map */
  for(i = minLineNb; i < maxLineNb; i++)
    {
      y = map->yOrigin + ((float) minColNb)  * map->scale;

      for(j = minColNb; j < maxColNb; j++)
	{
/* 	  fprintf(stderr, "Position: x= %f, y= %f\n", x, y); */
	  /*-- Place the robot on x,y position --*/
	  /*-- Do not print the result each time --*/
	  p3dPlaceRobotOnMap(&config, p3dplanner, x, y, thetaDeg, map, 2);
	  y += map->scale;
	}
      x += map->scale;
    }

  /*** Final Time ***/
  clock_gettime(CLOCK_REALTIME, &tp);
  msec = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
  timesec1  = ((double)(msec))/1000;
  /******************/

  fprintf(stderr, "------------ Bilan -------------\n");
  fprintf(stderr, "Total time (s): %f\n", timesec1-timesec0 );
  fprintf(stderr, "For a %dx%d map, ie: %d positions\n", 
	  maxLineNb-minLineNb, maxColNb-minColNb, 
	  (maxLineNb-minLineNb)*(maxColNb-minColNb) );
  fprintf(stderr, "--------------------------------\n");


  return 0;
}




int 
p3dMap_2_ima1 (P3D_MAP*	p3dMap,
	       unsigned char*	pixel)
{
  int i, j;
  unsigned char*	currentPixel;

#if 0
  P3D_MAP_CELL*		p3dMapPixel;

  currentPixel = pixel;
  p3dMapPixel  = p3dMap->cells; 

  for (i = 0; i < p3dMap->nbLines; i++) { 
	  for (j = 0; j < p3dMap->nbCols; j++) { 
		  /*-- Display in Calife way --*/
		  *currentPixel = (unsigned char) (255 * p3dMap->cells[i][j].state); 
		  /*-- Swap to get same display as with pom/GDHE --*/
		  *currentPixel = (unsigned char) (255 * p3dMap->cells[p3dMap->nbLines-i-1][p3dMap->nbCols-j-1].state); 
		  currentPixel++; 
		  p3dMapPixel++; 
	  } 
  } 
#endif


  currentPixel = pixel;

  for (j = 0; j < p3dMap->nbCols; j++) {
    for (i = 0; i < p3dMap->nbLines; i++) {

      if (p3dMap->cells[i][j].state == P3D_MAP_CELL_FILLED) {
	/* Swap (cf dem) pour remettre le mnt dans le sens habituel */
 	currentPixel = (pixel + ((p3dMap->nbCols - j  - 1)* p3dMap->nbLines) + (p3dMap->nbLines - i - 1));
	
	*pixel = (unsigned char) (255 * p3dMap->cells[i][j].state);
      }
    }
  }

  return 0;
}



/**********************************************************************
 *
 * Ã partir d'une P3D_MAP, construit une ima8 qui a la mÃªme taille, dans laquelle
 * on met l'altitude ou l'Ã©cart type des cellules
 *
 **********************************************************************/

int
p3dmap_to_imaf (P3D_MAP*  p3dmap,
		float*	  q_image,
		int	  nblig,
		int	  nbcol,
		int	  infoType)
{
  int		i, j;
  /* P3D_MAP_CELL*	cell; */
  float*	pixel;

/*   if (dem->cells != CELLS_DETERMINEES) */
/*     return error (ERR_DEM_VIDE); */
  
  /* On initialise l'image Ã  0.0 */
  memset ((float*) q_image, 0, nblig * nbcol * sizeof (float));
  
  /* cell = p3dmap->cells; */
  pixel = q_image;

  for (j = 0; j < p3dmap->nbCols; j++) {
    for (i = 0; i < p3dmap->nbLines; i++) {

      /* cell = (p3dmap->cells + (i * nbcol) + j); */
      /* pixel = (q_image + (j * nblig) + i); */

      if (p3dmap->cells[i][j].state == P3D_MAP_CELL_FILLED) {
	/* Swap (cf dem) pour remettre le mnt dans le sens habituel */
 	pixel = (q_image + ((nbcol - j  - 1)* nblig) + (nblig - i - 1));

	switch (infoType) {
	case 0:
	  *pixel = p3dmap->cells[i][j].z;
	  break;
	case 1:
	  *pixel = p3dmap->cells[i][j].sigma_z;
	  break;
	case 2:
	  *pixel = p3dmap->cells[i][j].confidence;
	  break;
	default:
	  fprintf(stderr, "%s : Invalid format image (%d)\n\n", __func__, infoType);
	  return -1;
	  /* 	  break; */
	  /* Unreached */
	}
      }
      /* cell++; */
      /* pixel++; */
    }
  }
  
  /* On Ã©crit ce qui est codÃ© dans l'image */
  printf ("\n\tThe image represents ");
  switch (infoType) {
  case 0:
    printf ("the altitude of cells\n\n");
    break;
  case 1:
    printf ("standard deviation on cells altitude\n\n");
    break;
  case 2:
    printf ("the confidence associated to the cell\n\n");
    break;
  default:
	fprintf(stderr, "%s : Invalid format image (%d)\n\n", __func__, infoType);
	return -1;
/* 	  break; */
/* Unreached */
  }

  return 0;
}


/**********************************************************************
 *
 * Creation and allocation of a P3DPlanSolution
 *
 **********************************************************************/

P3DPlanSolution *
create_P3DPlanSolution(int depth, int nNodes)
{
	P3DPlanSolution * res = malloc(sizeof(*res));
	int i, j;
	if (res == NULL)
		return NULL;

	memset(res, 0, sizeof(*res));

	/* Allocate BestArc */

	res->bestArc = calloc(depth, sizeof(P3DArc));
	if (res->bestArc == NULL) {
		free(res);
		return NULL;
	}

	for (i = 0; i < depth; ++i)
	{
		res->bestArc[i].nodes = calloc(nNodes, sizeof(P3DNode));
		if (res->bestArc[i].nodes == NULL) {
			for (j = 0; j < i ; ++j) 
				free(res->bestArc[j].nodes);
			free(res->bestArc);
			free(res);
			return NULL;
		}
	}

	res->_maxdepth = depth;
	res->depth = 0;
	return res;
}

/**********************************************************************
 *
 * Destruction of a P3DPlanSolution
 *
 **********************************************************************/

void
destroy_P3DPlanSolution(P3DPlanSolution *sol)
{
	int i;

	if (sol == NULL)
		return;

	if (sol->bestArc != NULL) {
		for (i = 0; i < sol->_maxdepth; ++i) {
			free(sol->bestArc[i].nodes);
		}
		free(sol->bestArc);
	}
	free(sol);
	return;
}

/**********************************************************************
 *
 * Reset of a P3DPlanSolution
 *
 **********************************************************************/
void
reset_P3DPlanSolution(P3DPlanSolution *sol)
{
	if (sol == NULL)
		return;

	sol->is_valid = 0;
	sol->curv = 0.0;
	sol->d = 0.0;
	sol->cost = 0.0;

	sol->v = 0.0;
	sol->w = 0.0;
	sol->alpha = 0.0;

	sol->depth = 0;

	sol->average_time = 0.0;
	sol->ncalls = 0;
}

/*
 * sub_timeval
 *
 * Substract two timeval and returns the result in second
 */

double 
sub_timeval(struct timeval *begin, struct timeval *end)
{
	long sec = end->tv_sec - begin->tv_sec;
	long usec = end->tv_usec - begin->tv_usec;
	if (usec < 0) {
		sec --;
		usec += 1000000;
	}

	return (double)sec+(double)usec/1000000.0;
}

double
p3dGetStopDistance(double v, P3DPlanner * planner)
{
	double res;
	res =  ((v * v) / ( 2 * planner->robot->decel)) + 2 * p3dGetVehiculeSize(planner);
	fprintf(stderr, "v : %lf || stop_distance : %lf\n", v, res);
	return res;
}

double
max_speed(const P3DArc *a, const P3DPlanner *planner) 
{
	double max_know_distance;
	double max_speed_unknow;
	double max_speed_obstacle;
	double max_speed_possibe;

	if (a->first_unknow == 0) {
		max_speed_unknow = planner->robot->vMax;
	} else {
		if (a->first_unknow != 1) {
			max_know_distance = (a->first_unknow - 1) * a->ds;
		} else { /* XXX It is incorrect ... */
			max_know_distance = a->ds;
		}

		max_speed_unknow = sqrt (2 * planner->robot->decel * max_know_distance);
	}


	if (a->current == (a->n -1)) {
		max_speed_obstacle = planner->robot->vMax;
	} else {
		if (a->current != 1) {
			max_know_distance = (a->current - 1) * a->ds;
		} else { /* XXX It is incorrect ... */
			max_know_distance = a->ds;
		}

		max_speed_obstacle = sqrt (2 * planner->robot->decel * max_know_distance);
	}

	if (max_speed_unknow < max_speed_obstacle)
		max_speed_possibe = max_speed_unknow;
	else
		max_speed_possibe = max_speed_obstacle;

	return max_speed_possibe;
}


void
check_unknow(P3DRobotGeometry *geom,  P3DArc *a, int current_node)
{
	if (a->first_unknow > 0)
		return;

	if (a->nodes[current_node].conf.pUnk > geom->unkMax) {
		a->first_unknow = current_node;
	}
}

void
p3dComputeRobotMovement(double *px, double *py, double *ptheta, P3DPlanSolution
		*sol, P3DPlanner * planner)
{
	double x, y, theta;
	double average_dist;
	double x_new, y_new, theta_new;
	double ds, curv;

	x = *px;
	y = *py;
	theta = *ptheta;
	/*
	 * For Trooper, we only can rely on x,y,theta, v, alpha
	 * Other information may be different than the one we have computed in p3d 
	 * (as we may not be the elected algo)
	 */
#if (defined(TROOPER))
	ds = 0.2;
	curv = sol->alpha / p3dGetDistanceFixAndMobileAxle(planner);
#else
	ds = sol->bestArc->ds;
	curv = sol->bestArc->r;
#endif

	/* 
	 * On first call, we can't compute the new position of the robot but 
	 * it is not a problem because normally our speed is 0 at the beginning
	 * of the world
	 */

	if (sol->ncalls) { 
		average_dist = sol->v * sol->average_time;

		while (average_dist > ds) {
			average_dist -= ds;
			p3dGetArcPos(curv, ds, x, y, theta, &x_new, &y_new, &theta_new);
			x = x_new;
			y = y_new;
			theta = theta_new;
		}

		/* the last case if average_dist <= ds. We compute the arc and 
		 * we compute the approximal position using a linear approx */

		p3dGetArcPos(curv, ds, x, y, theta, &x_new, &y_new, &theta_new);
		x = (x_new - x) * (average_dist / ds) + x;
		y = (y_new - y) * (average_dist / ds) + y ;
		theta = (theta_new - theta) * (average_dist / ds) + theta;
	}

	*px = x;
	*py = y;
	*ptheta = theta;
}

void
p3dComputeRobotPositionInMap(P3D_MAP * map, T3D *mainToOrigin, T3D *mainToCenterOfMap, 
		T3D *goalToOrigin, T3D *goalToCenterOfMap)
{
	T3D tmp;
	t3dInit(mainToCenterOfMap, T3D_EULER, T3D_ALLOW_CONVERSION);
	/*---- compute robot position in map ----*/
	/* tmp = inv(map->mainToOrigin) = originToMain(map) */
	t3dCopyIn(&tmp, map->mainToOrigin);
	tmp.flags |= T3D_ALLOW_CONVERSION;
	t3dInvert(&tmp);

	if (mainToOrigin != NULL && mainToCenterOfMap != NULL) { 
		t3dInit(mainToCenterOfMap, T3D_EULER, T3D_ALLOW_CONVERSION);

		mainToOrigin->flags |= T3D_ALLOW_CONVERSION;
		/* mainToCenterOfMap = comp(robotToOrigin,originToMap) = robotToMap */
		t3dCompIn(mainToCenterOfMap, mainToOrigin, &tmp);
		/* t3dComp(&mainToCenterOfMap, map->mainToCenter); */
		t3dConvertTo(T3D_EULER, mainToCenterOfMap);
	}


	if (goalToOrigin != NULL && goalToCenterOfMap != NULL) {
		t3dInit(goalToCenterOfMap, T3D_EULER, T3D_ALLOW_CONVERSION);

		goalToOrigin->flags |= T3D_ALLOW_CONVERSION;
		/*---- compute goal position ----*/
		/* goalToCenterOfMap = comp(goalToOrigin, originToMain(map) = goalToMain(map)*/
		t3dCompIn(goalToCenterOfMap, goalToOrigin, &tmp);
		/* t3dComp(&goalToCenterOfMap, map->mainToCenter); */
		t3dConvertTo(T3D_EULER, goalToCenterOfMap);
	}
}

/*
 * XXX It is not the real size of the vehicule, but it is ok for what we want to do
 * aka used as a marge for p3d. 
 *
 * If we really want the vehicule size, we need more information in the robot geometry
 */
double
p3dGetVehiculeSize(P3DPlanner *planner)
{
	return planner->geom->arm[0] + planner->geom->arm[1];
}

double
p3dGetDistanceFixAndMobileAxle(P3DPlanner *planner)
{
#if defined(TROOPER)
	return planner->geom->arm[0];
#elif (defined(ATRV) || defined(RMP400))
	return planner->geom->arm[0] + planner->geom->arm[1];
#else
#error "p3dGetDistanceFixAndMobileAxle is not defined for this kind of robot"
#endif
}


   	
